import argparse

from ctc_score import StyleTransferScorer, SummarizationScorer, DialogScorer
from ctc_score.configs import TASKS, ALIGNS


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument('--task', required=True, choices=TASKS)
    parser.add_argument('--align', required=True, choices=ALIGNS,
                        help='the align model to use')
    parser.add_argument('--aspect', required=True,
                        help='the aspect to evaluate')
    parser.add_argument('--hypo', required=True,
                        help='a file with all hypothesized texts to evaluate, '
                             'line-by-line')
    parser.add_argument('--remove_stopwords',
                        default=False, action='store_true',
                        help='whether to remove stopwords in aligning')
    parser.add_argument('--scores_save_path', default='/tmp/scores.txt',
                        help='a path to save example-wise scores.')

    # Dialog
    parser.add_argument('--fact', help='a file with all facts, line-by-line')
    parser.add_argument('--dialog_history',
                        help='a file with all dialog histories, line-by-line')

    # Summarization
    parser.add_argument('--doc', help='a file with all documents, line-by-line')
    parser.add_argument(
        '--refs',
        help='a file with all references, line-by-line. '
             'if each document has more than one reference, '
             'divide them by \"|||\"')

    # Style Transfer
    parser.add_argument('--input_sent',
                        help='a file with all input sentences, line-by-line')

    return parser.parse_args()


def evaluate_style_transfer(args):
    scorer = StyleTransferScorer(align=args.align)

    scores = []
    for input_sent, hypo in zip(
            open(args.input_text).readlines(),
            open(args.hypo).readlines()):
        input_sent, hypo = input_sent.strip(), hypo.strip()

        if input_sent == '' and hypo == '':
            continue
        scores.append(scorer.score(
            input_sent=input_sent,
            hypo=hypo,
            aspect=args.aspect,
            remove_stopwords=args.remove_stopwords))

    return scores


def evaluate_summarization(args):
    scorer = SummarizationScorer(align=args.align)

    scores = []
    for doc, refs, hypo in zip(
            open(args.doc).readlines(),
            open(args.refs).readlines(),
            open(args.hypo).readlines()):
        doc, refs, hypo = doc.strip(), refs.strip().split('|||'), hypo.strip()

        if doc == '' and hypo == '':
            continue
        scores.append(scorer.score(
            doc=doc,
            refs=refs,
            hypo=hypo,
            aspect=args.aspect,
            remove_stopwords=args.remove_stopwords))

    return scores


def evaluate_dialog(args):
    scorer = DialogScorer(align=args.align)

    scores = []
    for fact, dialog_history, hypo in zip(
            open(args.fact).readlines(),
            open(args.dialog_history).readlines(),
            open(args.hypo).readlines()):
        fact, dialog_history, hypo = \
            fact.strip(), dialog_history.strip(), hypo.strip()

        if fact == '' and dialog_history == '' and hypo == '':
            continue
        scores.append(scorer.score(
            fact=fact,
            dialog_history=dialog_history,
            hypo=hypo,
            aspect=args.aspect,
            remove_stopwords=args.remove_stopwords))

    return scores


def main():
    args = parse_args()

    if args.task == 'style_transfer':
        scores = evaluate_style_transfer(args)
    elif args.task == 'summarization':
        scores = evaluate_summarization(args)
    elif args.task == 'dialog':
        scores = evaluate_dialog(args)
    else:
        raise ValueError(f'\"task\" should be one of {TASKS}.')

    scores_file = open(args.scores_save_path, 'w')
    for score in scores:
        print(score, file=scores_file)

    print(f'#examples detected: {len(scores)}')
    print(f'{args.aspect} score: {sum(scores) / len(scores)}')
    print(f'example-wise scores are saved in {args.scores_save_path}')


if __name__ == '__main__':
    main()


from ctc_score.scorer import Scorer

from ctc_score.style_transfer_scorer import StyleTransferScorer
from ctc_score.summarization_scorer import SummarizationScorer
from ctc_score.dialog_scorer import DialogScorer
from ctc_score.download import maybe_download



from collections import Counter
import os
from typing import List

import torch
from torch import nn

from ctc_score.models.aligner import Aligner

from transformers import AdamW, get_linear_schedule_with_warmup, AutoTokenizer
from transformers import RobertaConfig, RobertaModel, RobertaTokenizerFast

from ctc_score.data_utils.data_utils import TokenClassificationExample, \
    text_clean, get_words


INIT = 'roberta.large'
MAX_LENGTH = 512


class DiscriminativeAligner(Aligner, nn.Module):
    def __init__(self, aggr_type):
        Aligner.__init__(self, aggr_type)
        nn.Module.__init__(self)
        self._roberta = RobertaModel.from_pretrained('roberta-large')
        self._roberta_tokenizer = RobertaTokenizerFast.from_pretrained(
            'roberta-large')

        self._classifier = nn.Linear(
            self._roberta.config.hidden_size, 2)
        self._tokenizer = AutoTokenizer.from_pretrained("roberta-large")

        self._hparams = None

    def forward(self, input_text, words, context):
        if len(self._tokenizer(input_text)['input_ids']) > MAX_LENGTH:
            print('Length of input text exceeds max length! Skipping')
            return None

        tokens = self._roberta_tokenizer(
            input_text, context, return_tensors='pt').to('cuda')
        tokens = {'input_ids': tokens['input_ids'][:, :MAX_LENGTH],
                  'attention_mask': tokens['attention_mask'][:, :MAX_LENGTH]}
        features = self._roberta(**tokens).last_hidden_state[0]

        try:
            len_input_text = len(
                self._roberta_tokenizer(input_text)['input_ids'])
            word_features = self.extract_features_aligned_to_words(
                input_text=input_text,
                words=words,
                features=features[:len_input_text])
        except:
            print(f'Error bpe-to-words, '
                  f'word_logits=None for this batch: {input_text}')
            return None

        word_logits = self._classifier(word_features)

        return word_logits

    def extract_features_aligned_to_words(self, input_text, words, features):
        bpe_toks = self._roberta_tokenizer(
            input_text, return_tensors='pt')['input_ids'][0]

        alignment = self.align_bpe_to_words(bpe_toks, words)
        aligned_feats = self.align_features_to_words(features, alignment)

        assert len(words) + 2 == len(aligned_feats), 'words length {} does not match feature lenth {}. '.format(
            len(words), len(aligned_feats))

        return aligned_feats[1:-1]

    def align_features_to_words(self, features, alignment):
        """
        Align given features to words.

        Args:
            roberta (RobertaHubInterface): RoBERTa instance
            features (torch.Tensor): features to align of shape `(T_bpe x C)`
            alignment: alignment between BPE tokens and words returned by
                func:`align_bpe_to_words`.
        """
        assert features.dim() == 2

        bpe_counts = Counter(
            j for bpe_indices in alignment for j in bpe_indices)
        assert bpe_counts[0] == 0  # <s> shouldn't be aligned
        denom = features.new([bpe_counts.get(j, 1)
                             for j in range(len(features))])
        weighted_features = features / denom.unsqueeze(-1)

        output = [weighted_features[0]]
        largest_j = -1
        for bpe_indices in alignment:
            output.append(weighted_features[bpe_indices].sum(dim=0))
            largest_j = max(largest_j, *bpe_indices)
        for j in range(largest_j + 1, len(features)):
            output.append(weighted_features[j])
        output = torch.stack(output)
        # assert torch.all(torch.abs(output.sum(dim=0) - features.sum(dim=0)) < 1e-4)
        return output

    def align_bpe_to_words(self, bpe_tokens: torch.LongTensor, other_tokens: List[str]):
        """
        Helper to align GPT-2 BPE to other tokenization formats (e.g., spaCy).

        Args:
            roberta (RobertaHubInterface): RoBERTa instance
            bpe_tokens (torch.LongTensor): GPT-2 BPE tokens of shape `(T_bpe)`
            other_tokens (List[str]): other tokens of shape `(T_words)`

        Returns:
            List[str]: mapping from *other_tokens* to corresponding *bpe_tokens*.
        """
        assert bpe_tokens.dim() == 1
        assert bpe_tokens[0] == 0

        def clean(text):
            return text.strip()

        # remove whitespaces to simplify alignment
        bpe_tokens = [self._roberta_tokenizer.decode([x]) for x in bpe_tokens]
        bpe_tokens = [clean(tok) if tok not in ['<s>', '</s>']
                      else '' for tok in bpe_tokens]

        other_tokens = [clean(str(o)) for o in other_tokens]

        # strip leading <s>
        bpe_tokens = bpe_tokens[1:]
        assert "".join(bpe_tokens) == "".join(other_tokens), 'bpe tokens:{}, other tokens:{}'.format(
            ''.join(bpe_tokens), ''.join(other_tokens))

        # create alignment from every word to a list of BPE tokens
        alignment = []
        bpe_toks = filter(
            lambda item: item[1] != "", enumerate(bpe_tokens, start=1))
        j, bpe_tok = next(bpe_toks)
        for other_tok in other_tokens:
            bpe_indices = []
            while True:
                if other_tok.startswith(bpe_tok):
                    bpe_indices.append(j)
                    other_tok = other_tok[len(bpe_tok):]
                    try:
                        j, bpe_tok = next(bpe_toks)
                    except StopIteration:
                        j, bpe_tok = None, None
                elif bpe_tok.startswith(other_tok):
                    # other_tok spans multiple BPE tokens
                    bpe_indices.append(j)
                    bpe_tok = bpe_tok[len(other_tok):]
                    other_tok = ""
                else:
                    raise Exception(
                        'Cannot align "{}" and "{}"'.format(other_tok, bpe_tok))
                if other_tok == "":
                    break
            assert len(bpe_indices) > 0
            alignment.append(bpe_indices)
        assert len(alignment) == len(other_tokens)

        return alignment

    def align(self, input_text, context):
        input_text, context = text_clean(input_text), text_clean(context)

        input_text = self._roberta_tokenizer.decode(
            self._roberta_tokenizer(input_text, return_tensors='pt')['input_ids'][0][:MAX_LENGTH // 2])

        input_text = input_text.replace('<s>', '').replace('</s>', '')
        word_logits = self(
            input_text=input_text,
            words=get_words(input_text),
            context=context)

        if word_logits is None:
            return None, None

        preds = torch.softmax(word_logits, dim=-1)[:, 0]

        return get_words(input_text), preds.tolist()



from ctc_score.models.aligner import Aligner
from bleurt.score import BleurtScorer

class BLEURTAligner(Aligner, BleurtScorer): 
    def __init__(self, aggr_type, *args, **kwargs):
        Aligner.__init__(self, aggr_type=None)
        BleurtScorer.__init__(self, *args, **kwargs)
        
    def align(self, input_text, context):
        score = self.score(references=[context], candidates=[input_text])
        tokens = input_text.split()
        return tokens, score


from ctc_score.scorer import Scorer


class StyleTransferScorer(Scorer):
    def __init__(self, align, aggr_type='mean', device='cuda'):
        Scorer.__init__(self, align=align, aggr_type=aggr_type, device=device)

    def score(self, input_sent, hypo, aspect, remove_stopwords=False):
        kwargs = dict(
            input_sent=input_sent,
            hypo=hypo,
            remove_stopwords=remove_stopwords)

        if aspect == 'preservation':
            return self.score_preservation(**kwargs)
        else:
            raise NotImplementedError

    def score_preservation(self, input_sent, hypo, remove_stopwords):
        aligner = self._get_aligner('sent_to_sent')

        align_y_x = aligner.get_score(
            input_text=input_sent,
            context=hypo,
            remove_stopwords=remove_stopwords)

        align_x_y = aligner.get_score(
            input_text=hypo,
            context=input_sent,
            remove_stopwords=remove_stopwords)

        return (align_y_x * align_x_y) / (align_y_x + align_x_y)




from nltk.corpus import stopwords


en_stopwords = stopwords.words('english')


class Aligner:
    def __init__(self, aggr_type):
        self._aggr_type = aggr_type

    def align(self, input_text, context):
        raise NotImplementedError

    def aggregate(self, tokens, token_scores, remove_stopwords):
        if self._aggr_type is None: 
            return token_scores[0]
        
        assert len(tokens) == len(token_scores)

        scores_to_aggr = []
        for token, score in zip(tokens, token_scores):
            if (token.lower() not in en_stopwords) or (not remove_stopwords):
                scores_to_aggr.append(score)

        if self._aggr_type == 'mean':
            return sum(scores_to_aggr) / len(scores_to_aggr) \
                if len(scores_to_aggr) > 0 else 0.
        elif self._aggr_type == 'sum':
            return sum(scores_to_aggr)
        else:
            raise ValueError

    def get_score(self, input_text, context, remove_stopwords):
        tokens, token_scores = self.align(
            input_text=input_text, context=context)

        if tokens is None:
            return None

        return self.aggregate(
            tokens=tokens,
            token_scores=token_scores,
            remove_stopwords=remove_stopwords)